/* Copyright 2011-2014 Yorba Foundation
 *
 * This software is licensed under the GNU Lesser General Public License
 * (version 2.1 or later).  See the COPYING file in this distribution.
 */

public errordomain Geary.EngineError {
    OPEN_REQUIRED,
    ALREADY_OPEN,
    ALREADY_EXISTS,
    NOT_FOUND,
    BAD_PARAMETERS,
    BAD_RESPONSE,
    INCOMPLETE_MESSAGE,
    SERVER_UNAVAILABLE,
    ALREADY_CLOSED,
    CLOSE_REQUIRED,
    CORRUPT,
    PERMISSIONS,
    VERSION,
    UNSUPPORTED
}

