/* Copyright 2013-2014 Yorba Foundation
 *
 * This software is licensed under the GNU Lesser General Public License
 * (version 2.1 or later).  See the COPYING file in this distribution.
 */

/**
 * Errors related to {@link Geary.Mime}.
 */

public errordomain MimeError {
    PARSE
}
