/* Copyright 2011-2014 Yorba Foundation
 *
 * This software is licensed under the GNU Lesser General Public License
 * (version 2.1 or later).  See the COPYING file in this distribution.
 */

public errordomain Geary.RFC822Error {
    INVALID,
    NOT_FOUND,
    FAILED
}

